# coding: utf8
from __future__ import absolute_import, division


import unittest as ut
import io

try:
    import xml.etree.cElementTree as et
except ImportError:
    import xml.etree.ElementTree as et

from plwn.utils.graphmlout import GraphMLWordNet


__all__ = 'GraphMLTest',


class GraphMLTest(ut.TestCase):

    def setUp(self):
        self.__gmo = GraphMLWordNet()

    def ensure_directed(self):
        """Graph type should always be directed."""
        graph = self.__write_and_read()
        self.assertEqual(
            graph.find(u'./graph').attrib[u'edgedefault'],
            u'directed',
        )

    def test_node(self):
        """Test adding a node."""
        self.__gmo.add_node(u'foo')
        result = self.__write_and_read()
        nodes = result.findall(u'./graph/node')

        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].attrib[u'id'], u'foo')

    def test_edge(self):
        """Test adding an edge."""
        self.__gmo.add_edge(u'a-b-c', u'a', u'b')
        result = self.__write_and_read()
        edges = result.findall(u'./graph/edge')

        self.assertEqual(len(edges), 1)
        self.assertEqual(edges[0].attrib[u'id'], u'a-b-c')
        self.assertEqual(edges[0].attrib[u'source'], u'a')
        self.assertEqual(edges[0].attrib[u'target'], u'b')

    def test_node_badattr(self):
        """Add an attribute, which was not defined, to a node."""
        self.assertRaises(
            KeyError,
            self.__gmo.add_node,
            u'foo',
            {u'bar': u'baz'},
        )

    def test_edge_badattr(self):
        """Add an attribute, which was not defined, to an edge."""
        self.assertRaises(
            KeyError,
            self.__gmo.add_edge,
            u'foo-bar',
            u'foo',
            u'bar',
            {u'bar': u'baz'},
        )

    def test_str_attribute(self):
        """Add a string attribute and make sure the key is added."""
        self.__gmo.add_attribute_type(
            u'foo',
            u'foo_param',
            GraphMLWordNet.DATA_TYPE_STR,
        )
        # Add the node to an attribute
        self.__gmo.add_node(u'oof', {u'foo': u'ala'})

        result = self.__write_and_read()
        keys = result.findall(u'./key')

        self.assertEqual(len(keys), 1)
        self.assertEqual(keys[0].attrib[u'id'], u'foo')
        self.assertEqual(keys[0].attrib[u'attr.name'], u'foo_param')
        self.assertEqual(keys[0].attrib[u'attr.type'], u'string')
        self.assertEqual(keys[0].attrib[u'for'], u'node')

        nodedata = result.findall(u"./graph/node[@id='oof']/data")

        self.assertEqual(len(nodedata), 1)
        self.assertEqual(nodedata[0].attrib[u'key'], u'foo')
        self.assertEqual(nodedata[0].text, u'ala')

    def test_int_attribute(self):
        """Add an int attribute and make sure the key is added."""
        self.__gmo.add_attribute_type(
            u'foo',
            u'foo_param',
            GraphMLWordNet.DATA_TYPE_INT,
        )
        # Add the node to an attribute
        self.__gmo.add_node(u'oof', {u'foo': 1})

        result = self.__write_and_read()
        keys = result.findall(u'./key')

        self.assertEqual(len(keys), 1)
        self.assertEqual(keys[0].attrib[u'id'], u'foo')
        self.assertEqual(keys[0].attrib[u'attr.name'], u'foo_param')
        self.assertEqual(keys[0].attrib[u'attr.type'], u'long')
        self.assertEqual(keys[0].attrib[u'for'], u'node')

        nodedata = result.findall(u"./graph/node[@id='oof']/data")

        self.assertEqual(len(nodedata), 1)
        self.assertEqual(nodedata[0].attrib[u'key'], u'foo')
        self.assertEqual(nodedata[0].text, u'1')

    def test_bool_attribute(self):
        """Add a boolean attribute and make sure the key is added."""
        self.__gmo.add_attribute_type(
            u'foo',
            u'foo_param',
            GraphMLWordNet.DATA_TYPE_BOOL,
        )
        # Add the node to an attribute
        self.__gmo.add_node(u'oof', {u'foo': True})
        self.__gmo.add_node(u'boof', {u'foo': False})

        result = self.__write_and_read()
        keys = result.findall(u'./key')

        self.assertEqual(len(keys), 1)
        self.assertEqual(keys[0].attrib[u'id'], u'foo')
        self.assertEqual(keys[0].attrib[u'attr.name'], u'foo_param')
        self.assertEqual(keys[0].attrib[u'attr.type'], u'boolean')
        self.assertEqual(keys[0].attrib[u'for'], u'node')

        nodedata = result.findall(u"./graph/node[@id='oof']/data")

        self.assertEqual(len(nodedata), 1)
        self.assertEqual(nodedata[0].attrib[u'key'], u'foo')
        self.assertEqual(nodedata[0].text, u'true')

        nodedata2 = result.findall(u"./graph/node[@id='boof']/data")

        self.assertEqual(len(nodedata2), 1)
        self.assertEqual(nodedata2[0].attrib[u'key'], u'foo')
        self.assertEqual(nodedata2[0].text, u'false')

    def test_json_attribute(self):
        """Add a JSON attribute and make sure the key is added."""
        self.__gmo.add_attribute_type(
            u'foo',
            u'foo_param',
            GraphMLWordNet.DATA_TYPE_JSON,
        )
        # Add the node to an attribute
        self.__gmo.add_node(u'oof', {u'foo': (1, 2, u'3')})

        result = self.__write_and_read()
        keys = result.findall(u'./key')

        self.assertEqual(len(keys), 1)
        self.assertEqual(keys[0].attrib[u'id'], u'foo')
        self.assertEqual(keys[0].attrib[u'attr.name'], u'foo_param')
        self.assertEqual(keys[0].attrib[u'attr.type'], u'string')
        self.assertEqual(keys[0].attrib[u'for'], u'node')

        nodedata = result.findall(u"./graph/node[@id='oof']/data")

        self.assertEqual(len(nodedata), 1)
        self.assertEqual(nodedata[0].attrib[u'key'], u'foo')
        self.assertEqual(nodedata[0].text, u'[1, 2, "3"]')

    # TODO: No enum tests

    def test_edge_attribute(self):
        """Add an ordinary string attribute, but to an edge."""
        self.__gmo.add_attribute_type(
            u'foo',
            u'foo_param',
            GraphMLWordNet.DATA_TYPE_STR,
            u'edge',
        )
        # Add an edge for this attribute
        self.__gmo.add_edge(u'a-b', u'a', u'b', {u'foo': u'bar'})

        result = self.__write_and_read()
        edgedata = result.findall(u"./graph/edge[@id='a-b']/data")

        self.assertEqual(len(edgedata), 1)
        self.assertEqual(edgedata[0].attrib[u'key'], u'foo')
        self.assertEqual(edgedata[0].text, u'bar')

    def __write_and_read(self):
        with io.BytesIO() as ss:
            self.__gmo.write(ss)
            return et.fromstring(ss.getvalue())
