# -*- coding: utf-8 -*-

# FIXME Comments down below suggest that there are compatibility issues with
# some tests. Fix and uncomment them.

from __future__ import absolute_import, division


import unittest
from plwn.readers import ubylmf as ur
from plwn.readers.nodes import make_synset_node, make_lexical_unit_node
from plwn.enums import PoS, Domain
from plwn import exceptions as exc
from io import BytesIO
from xml.etree import ElementTree as et
# Python2
try:
    from itertools import izip
# Python3
except ImportError:
    izip = zip

__all__ = 'UBYLMFReaderTest',

ENCODING = 'UTF-8'

test_xml = u"""<?xml version="1.0" encoding="UTF-8" ?>
<LexicalResource dtdVersion="ubyDTD_1_0.dtd" name="plWordnet">
<Lexicon languageIdentifier="pl" id="1" name="Słowosieć 2.2">

<LexicalEntry id="15" partOfSpeech="rzeczownik">
    <Lemma>
        <FormRepresentation writtenForm="'patafizyka"/>
    </Lemma>
    <Sense id="plWN_Sense_628506" index="1" synset="plWN_Synset_396603">
        <Definition>
            <Statement statementType="usageNote">
                <TextRepresentation writtenText="specj."/>
            </Statement>
            <Statement statementType="externalReference">
                <TextRepresentation
                    writtenText="http://pl.wikipedia.org/wiki/Patafizyka"/>
            </Statement>
        </Definition>
        <MonolingualExternalRef
            externalSystem="WordnetLoom PartOfSpeech And SenseIndex"
            externalReference="POS[plWN rzeczownik] 0"/>
        <SemanticLabel label="noun.plWN_umy" type="domain"/>
    </Sense>
    <Sense id="plWN_Sense_54584" index="2" synset="plWN_Synset_36078">
        <SenseRelation target="plWN_Sense_17308" relName="deminutywność"/>
        <MonolingualExternalRef
            externalSystem="WordnetLoom PartOfSpeech And SenseIndex"
            externalReference="POS[plWN rzeczownik] 0"/>
        <SemanticLabel label="noun.plWN_wytw" type="domain"/>
    </Sense>
    <Sense id="plWN_Sense_54584" index="3" synset="plWN_Synset_!#!">
        <SenseRelation target="plWN_Sense_17308" relName="deminutywność"/>
        <MonolingualExternalRef
            externalSystem="WordnetLoom PartOfSpeech And SenseIndex"
            externalReference="POS[plWN rzeczownik] 0"/>
        <SemanticLabel label="noun.plWN_wytw" type="domain"/>
    </Sense>
</LexicalEntry>
<Synset id="plWN_Synset_10">
    <SynsetRelation target="plWN_Synset_9139" relName="hiperonimia"/>
    <SynsetRelation target="plWN_Synset_19032" relName="wartość_cechy"/>
    <SynsetRelation target="plWN_Synset_104177" relName="hiperonimia"/>
    <SynsetRelation target="plWN_Synset_105404" relName="hiperonimia"/>
    <SynsetRelation target="plWN_Synset_228433" relName="hiperonimia"/>
</Synset>
<Synset id="plWN_Synset_246792">
    <SynsetRelation target="plWN_Synset_245829" relName="hiponimia"/>
</Synset>
</Lexicon>
<SenseAxis id="433581" synsetOne="plWN_Synset_246792"
    synsetTwo="WordNet 3.1 synset offset: 4730898">
    <SenseAxisRelation target="433581" relName="Hipo_plWN-PWN"/>
</SenseAxis>
</LexicalResource>
"""


class UBYLMFReaderTest(unittest.TestCase):
    def test_ubylmf_reader(self):
        filelike = BytesIO(b"")
        self.assertRaises(
            et.ParseError,
            next,
            ur.ubylmf_reader(filelike),
        )

        lu1 = make_lexical_unit_node(
            id=628506,
            lemma=u"'patafizyka",
            pos=PoS.n,
            synset=396603,
            unit_index=0,
            usage_notes=(u"specj.",),
            external_links=(u"http://pl.wikipedia.org/wiki/Patafizyka",),
            domain=Domain.umy,
            variant=1,
        )
        lu2 = make_lexical_unit_node(
            id=54584,
            lemma=u"'patafizyka",
            pos=PoS.n,
            synset=36078,
            unit_index=0,
            domain=Domain.wytw,
            related=((u"deminutywność", 17308), ),
            variant=2,
        )
        s1 = make_synset_node(
            id=10,
            related=(
                (u"hiperonimia", 9139),
                (u"wartość_cechy", 19032),
                (u"hiperonimia", 104177),
                (u"hiperonimia", 105404),
                (u"hiperonimia", 228433)
            ),
        )
        s2 = make_synset_node(
            id=246792,
            related=((u"hiponimia", 245829),),
        )
        iter_corr = iter((lu1, lu2, s1, s2))
        filelike = BytesIO(test_xml.encode('UTF-8'))
        for (read, correct) in izip(ur.ubylmf_reader(filelike), iter_corr):
            self.assertEqual(read, correct)

    def test_make_lexicalunit(self):
        # Wrong sense's ID
        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584?" index="3" synset="plWN_Synset_!#!">
            </Sense>
            """.encode(ENCODING)
        )
        self.assertRaises(
            exc.MalformedIdentifierException,
            ur._make_lexicalunit,
            et.Element(u"LexicalEntry"),
            xml_sense
        )
        # Incorrect synset's ID
        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="3" synset="plWN_Synset_!#!">
            </Sense>
            """.encode(ENCODING)
        )
        self.assertRaises(
            exc.MalformedIdentifierException,
            ur._make_lexicalunit,
            et.Element(u"LexicalEntry"),
            xml_sense
        )
        # Incorrect variant (index)
        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="A" synset="plWN_Synset_33">
            </Sense>
            """.encode(ENCODING)
        )
        self.assertRaises(
            ValueError,
            ur._make_lexicalunit,
            et.Element(u"LexicalEntry"),
            xml_sense
        )
        # Missing <Lemma>
        xml_lu = et.fromstring(
            u"""
            <LexicalEntry id="15" partOfSpeech="rzeczownik">
            </LexicalEntry>
            """.encode(ENCODING)
        )
        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="0" synset="plWN_Synset_33">
            </Sense>
            """.encode(ENCODING)
        )
        self.assertRaises(
            AttributeError,
            ur._make_lexicalunit,
            xml_lu,
            xml_sense
        )
        # Empty <Lemma>
        xml_lu = et.fromstring(
            u"""
            <LexicalEntry id="15" partOfSpeech="rzeczownik">
                <Lemma>
                    <FormRepresentation writtenForm=""/>
                </Lemma>
            </LexicalEntry>
            """.encode(ENCODING)
        )
        self.assertRaisesRegexp(
            AssertionError,
            "Lemma is empty",
            ur._make_lexicalunit,
            xml_lu,
            xml_sense
        )
        # Empty PoS
        xml_lu = et.fromstring(
            u"""
            <LexicalEntry id="15" partOfSpeech="">
                <Lemma>
                    <FormRepresentation writtenForm="'patafizyka"/>
                </Lemma>
            </LexicalEntry>
            """.encode(ENCODING)
        )
        self.assertRaisesRegexp(
            AssertionError,
            "PoS is empty",
            ur._make_lexicalunit,
            xml_lu,
            xml_sense
        )
        # Don't check definitions and domain - they are in separate tests
        # Incorrect unit index
        xml_lu = et.fromstring(
            u"""
            <LexicalEntry id="15" partOfSpeech="rzeczownik">
                <Lemma>
                    <FormRepresentation writtenForm="'patafizyka"/>
                </Lemma>
            </LexicalEntry>
            """.encode(ENCODING)
        )
        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="0" synset="plWN_Synset_33">
                <SemanticLabel label="noun.plWN_wytw" type="domain"/>
                <MonolingualExternalRef
                    externalSystem="WordnetLoom PartOfSpeech And SenseIndex"
                    externalReference="POS[plWN rzeczownik] A"/>
            </Sense>
            """.encode(ENCODING)
        )
        self.assertRaises(
            exc.MalformedIdentifierException,
            ur._make_lexicalunit,
            xml_lu,
            xml_sense
        )
        # Check LU
        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_628506" index="1"
            synset="plWN_Synset_396603">
                <Definition>
                    <TextRepresentation writtenText="ExampleDefinition"/>
                </Definition>
                <Definition>
                    <Statement statementType="usageNote">
                        <TextRepresentation writtenText="specj."/>
                    </Statement>
                    <Statement statementType="usageNote">
                        <TextRepresentation writtenText="specj1"/>
                    </Statement>
                    <Statement statementType="externalReference">
                        <TextRepresentation writtenText="http://"/>
                    </Statement>
                </Definition>
                <MonolingualExternalRef
                    externalSystem="WordnetLoom PartOfSpeech And SenseIndex"
                    externalReference="POS[plWN rzeczownik] 0"/>
                <SemanticLabel label="noun.plWN_umy" type="domain"/>
                <SenseRelation target="plWN_Sense_17308"
                    relName="deminutywność"/>
                <SenseRelation target="plWN_Sense_17309"
                    relName="deminutywność"/>
                <SenseExample id="0">
                    <TextRepresentation
                        writtenText="1° = (π/180) rad = 60′ = 3600″[##W:]"/>
                </SenseExample>
                <SenseExample id="1">
                    <TextRepresentation writtenText="1[##P:]"/>
                </SenseExample>
            </Sense>
            """.encode(ENCODING)
        )
        lu = make_lexical_unit_node(
            id=628506,
            lemma=u"'patafizyka",
            pos=PoS.n,
            synset=396603,
            unit_index=0,
            definition=u"ExampleDefinition",
            usage_notes=(u"specj.", u"specj1",),
            external_links=("http://",),
            examples=(u"1° = (π/180) rad = 60′ = 3600″", u"1"),
            examples_sources=(u'W', u'P'),
            domain=Domain.umy,
            related=((u"deminutywność", 17308), (u"deminutywność", 17309)),
            variant=1,
        )
        self.assertEqual(lu, ur._make_lexicalunit(xml_lu, xml_sense))

    def test_make_synset(self):
        xml_synset = et.fromstring(
            u"""
            <Synset id="plWN_Synset_246792">
            </Synset>
            """.encode(ENCODING)
        )
        sn = make_synset_node(id=246792)
        self.assertEqual(sn, ur._make_synset(xml_synset))

        xml_synset = et.fromstring(
            u"""
            <Synset id="plWN_Synset_246792">
                <Definition>
                    <TextRepresentation writtenText="ExampleDefinition"/>
                </Definition>
            </Synset>
            """.encode(ENCODING)
        )
        sn = make_synset_node(id=246792, definition=u"ExampleDefinition")
        self.assertEqual(sn, ur._make_synset(xml_synset))

        xml_synset = et.fromstring(
            u"""
            <Synset id="plWN_Synset_246792">
                <SynsetRelation target="plWN_Synset_245829"
                    relName="hiponimia"/>
            </Synset>
            """.encode(ENCODING)
        )
        sn = make_synset_node(id=246792, related=((u"hiponimia", 245829),))
        self.assertEqual(sn, ur._make_synset(xml_synset))

        xml_synset = et.fromstring(
            u"""
            <Synset id="plWN_Synset_246792">
                <Definition>
                    <TextRepresentation writtenText="ExampleDefinition"/>
                </Definition>
                <SynsetRelation target="plWN_Synset_245829" relName="hip"/>
                <SynsetRelation target="plWN_Synset_245828" relName="ó"/>
            </Synset>
            """.encode(ENCODING)
        )
        sn = make_synset_node(
            id=246792,
            definition=u"ExampleDefinition",
            related=((u"hip", 245829), (u"ó", 245828)),
        )
        self.assertEqual(sn, ur._make_synset(xml_synset))

        xml_synset = et.fromstring(
            u"""
            <Synset id="plWN_Synset_??">
            </Synset>
            """.encode(ENCODING)
        )
        self.assertRaises(
            exc.MalformedIdentifierException,
            ur._make_synset,
            xml_synset
        )

        xml_synset = et.fromstring(
            u"""
            <Synset id="plWN_Synset_246792">
                <SynsetRelation target="plWN_Synset_245829" relName="hip"/>
                <SynsetRelation target="plWN_Synset_?" relName="hip"/>
            </Synset>
            """.encode(ENCODING)
        )
        sn = make_synset_node(id=246792, related=((u"hip", 245829),))
        self.assertEqual(sn, ur._make_synset(xml_synset))

    def test_extract_definitions(self):
        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="3" synset="plWN_Synset_!#!">
            </Sense>
            """.encode(ENCODING)
        )
        self.assertTupleEqual(
            (None, [], []),
            ur._extract_definitions(xml_sense)
        )

        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="3" synset="plWN_Synset_!#!">
                <Definition>
                    <TextRepresentation writtenText="ExampleDefinition"/>
                </Definition>
            </Sense>
            """.encode(ENCODING)
        )
        self.assertTupleEqual(
            (u"ExampleDefinition", [], []),
            ur._extract_definitions(xml_sense)
        )

        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="3" synset="plWN_Synset_!#!">
                <Definition>
                    <Statement statementType="?">
                        <TextRepresentation writtenText="specj."/>
                    </Statement>
                    <Statement statementType="externalReference">
                        <TextRepresentation writtenText="http://"/>
                    </Statement>
                    <Statement statementType="externalReference">
                        <TextRepresentation writtenText="http://1"/>
                    </Statement>
                </Definition>
            </Sense>
            """.encode(ENCODING)
        )
        self.assertTupleEqual(
            (None, [], [u"http://", u"http://1"]),
            ur._extract_definitions(xml_sense)
        )

        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="3" synset="plWN_Synset_!#!">
                <Definition>
                    <Statement statementType="usageNote">
                        <TextRepresentation writtenText="n1"/>
                    </Statement>
                    <Statement statementType="usageNote">
                        <TextRepresentation writtenText="łóż2"/>
                    </Statement>
                    <Statement statementType="?">
                        <TextRepresentation writtenText="specj."/>
                    </Statement>
                    <Statement statementType="externalReference">
                        <TextRepresentation writtenText="http://"/>
                    </Statement>
                    <Statement statementType="externalReference">
                        <TextRepresentation writtenText="http://1"/>
                    </Statement>
                </Definition>
            </Sense>
            """.encode(ENCODING)
        )
        self.assertTupleEqual(
            (None, [u"n1", u"łóż2"], [u"http://", u"http://1"]),
            ur._extract_definitions(xml_sense)
        )

        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="3" synset="plWN_Synset_!#!">
                <Definition>
                </Definition>
                <Definition>
                </Definition>
                <Definition>
                </Definition>
            </Sense>
            """.encode(ENCODING)
        )
        self.assertRaisesRegexp(
            AssertionError,
            r"Too many definitions \(3\)",
            ur._extract_definitions,
            xml_sense
        )

    def test_get_domain(self):
        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="3" synset="plWN_Synset_!#!">
                <SemanticLabel label="noun.plWN_wytw" type="domain"/>
            </Sense>
            """.encode(ENCODING)
        )
        self.assertEqual(u"noun.plWN_wytw", ur._get_domain(xml_sense))

        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="3" synset="plWN_Synset_!#!">
                <SemanticLabel label="ółźć" type="domain"/>
            </Sense>
            """.encode(ENCODING)
        )
        self.assertEqual(u"ółźć", ur._get_domain(xml_sense))

        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="3" synset="plWN_Synset_!#!">
                <SemanticLabel label="1" type="domain"/>
                <SemanticLabel label="2" type="domain"/>
            </Sense>
            """.encode(ENCODING)
        )
        self.assertRaises(AssertionError, ur._get_domain, xml_sense)
        self.assertRaisesRegexp(
            AssertionError,
            "2 SemanticLabel found, should be 1",
            ur._get_domain,
            xml_sense
        )

        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="3" synset="plWN_Synset_!#!">
                <SemanticLabel label="1" type=""/>
            </Sense>
            """.encode(ENCODING)
        )
        self.assertRaises(AssertionError, ur._get_domain, xml_sense)
        # Python3 fails here, so leave it... Unless you have an idea how to
        # make it in a cleaver way.
        # self.assertRaisesRegexp(
        #     AssertionError,
        #     "SemanticLabel has type  instead of domain",
        #     ur._get_domain,
        #     xml_sense
        # )

        xml_sense = et.fromstring(
            u"""
            <Sense id="plWN_Sense_54584" index="3" synset="plWN_Synset_!#!">
                <SemanticLabel label="1" type="ĄĘÓŁŻ"/>
            </Sense>
            """.encode(ENCODING)
        )
        self.assertRaises(AssertionError, ur._get_domain, xml_sense)
        # assertRaisesRegexp tries to convert unicode to str implicitly
        # without checking explicitly which version of Python is in use
        # this won't work...
        # self.assertRaisesRegexp(
        #     AssertionError,
        #     "SemanticLabel has type ĄĘÓŁŻ instead of domain",
        #     ur._get_domain,
        #     xml_sense
        # )

    def test_extract_id(self):
        self.assertEqual(ur._extract_id(u"id_1_and_id_2_234562"), 234562)
        self.assertEqual(ur._extract_id(u"234562"), 234562)
        self.assertEqual(ur._extract_id(u"łóżźć_234562"), 234562)
        self.assertRaises(
            exc.MalformedIdentifierException,
            ur._extract_id,
            u"id_"
        )
        self.assertRaises(
            exc.MalformedIdentifierException,
            ur._extract_id,
            u"łóżźć"
        )


if __name__ == '__main__':
    unittest.main()
