# coding: utf8
from __future__ import absolute_import, division, print_function

import itertools as itt
import logging
import unittest as ut
import sys

from plwn import enums as en
from plwn.readers import nodes as nd
from plwn.bases import RelationInfoBase

_IS_PY2 = sys.version_info.major == 2

if _IS_PY2:
    from plwn.readers.wndb import WNDBReader
    import wndbmockup as wndbm
else:
    wndbm = None
    WNDBReader = None


__all__ = (
    'GoodTest',
    'NonexistentRelationTest',
    'NonexistentSynsetInRelTest',
    'NonexistentUnitInRelTest',
    'SynsetWithNoUnitTest',
    'SynsetWithNoUnitInRelTest',
    'UnitWithNoSynsetInRelTest',
    'UnitWithNoSynsetTest',
    'UnitWithNonexistentSynsetTest',
)


# WNDBReader in general only works with python 2 (for now)
_py2_only = ut.skipIf(not _IS_PY2, 'Python 2 only functionality')

_NODE_ORDER = [nd.RelationTypeNode, nd.LexicalUnitNode, nd.SynsetNode]


@_py2_only
class GoodTest(ut.TestCase):
    """A case where everything's correct, touching all relevant tables."""

    def setUp(self):
        self.__dbm = wndbm.WnDbMockup()

        try:
            self.__setup_mock()
        except BaseException:
            self.__dbm.close()
            raise

    def __setup_mock(self):
        self.__dbm.add_synset(1, u'Duży pies')
        self.__dbm.add_synset(2, u'Duży ssak', abstract=1)
        self.__dbm.add_synset(3, u'Duży kundel')

        rel_hipo = self.__dbm.add_relation_type(u'hiponimia', u'hipo')
        # Make this relation have a fake parent, for testing
        rel_par = self.__dbm.add_relation_type(u'foo', u'f')
        rel_hiper = self.__dbm.add_relation_type(
            u'hiperonimia',
            u'hiper',
            parent=rel_par,
        )

        self.__dbm.add_synset_relation(1, 2, rel_hipo)
        self.__dbm.add_synset_relation(1, 3, rel_hiper)

        self.__dbm.add_lexical_unit(
            synid=1,
            lemma=u'psisko',
            pos=2,
            variant=1,
            unitindex=1,
            id_=11,
            domain=1,
            comment=u'##K: pot. [##W: Czarne psisko na rogu.]',
        )
        self.__dbm.add_lexical_unit(
            synid=1,
            lemma=u'suczysko',
            pos=2,
            variant=1,
            unitindex=2,
            id_=12,
            domain=1,
            comment=u'##D: Samica dużego psa',
        )
        self.__dbm.add_lexical_unit(
            synid=2,
            lemma=u'ssaczysko',
            pos=2,
            variant=1,
            unitindex=1,
            id_=21,
            domain=1,
        )
        self.__dbm.add_lexical_unit(
            synid=3,
            lemma=u'kundlisko',
            # Make it a verb even if it isn't, to test out
            pos=1,
            variant=1,
            unitindex=1,
            id_=31,
            domain=2,
            comment=u'{##L: www.mieszance.pl} {##L: dogpedia.com/mutt}',
            verb_aspect=4,
        )

        self.__dbm.add_emotion(
            lex_id=31,
            emotions=u'smutek;wstręt',
            valuations=None,
            markedness=u'- s',
            example1=u'bam',
            unit_status=1,
        )
        self.__dbm.add_emotion(
            lex_id=31,
            emotions=None,
            valuations=u'dobro;prawda;bob',
            markedness=u'-m',
            example2=u'bim',
            unit_status=1,
            super_annotation=1,
        )
        # Additional test for coalescing emotion data: emotions and valuations
        # must be uniq'd.
        self.__dbm.add_emotion(
            lex_id=31,
            emotions=u'radość;wstręt',
            valuations=u'piękno;prawda',
            markedness=u'+m',
            example1=u'uh',
            unit_status=1,
        )

        rel_zen = self.__dbm.add_relation_type(
            u'żeńskość',
            u'zen',
            is_syn=False,
        )

        self.__dbm.add_lexical_relation(11, 12, rel_zen)

    def tearDown(self):
        self.__dbm.close()

    def runTest(self):
        with self.__dbm.sqlalchemy_url_file as db_file:
            nodes = _process_read_nodes(self, WNDBReader(db_file))

        self.assertEqual(
            nodes,
            set((
                nd.make_relation_type_node(
                    kind=en.RelationKind.synset,
                    name=u'hiponimia',
                    aliases=(u'hipo',),
                ),
                nd.make_relation_type_node(
                    kind=en.RelationKind.synset,
                    name=u'hiperonimia',
                    aliases=(u'hiper',),
                    parent=u'foo',
                ),
                nd.make_relation_type_node(
                    kind=en.RelationKind.lexical,
                    name=u'żeńskość',
                    aliases=(u'zen',),
                ),
                nd.make_synset_node(
                    id=1,
                    definition=u'Duży pies',
                    related=(
                        (u'hiponimia', 2),
                        (
                            RelationInfoBase.format_name(
                                u'foo',
                                u'hiperonimia',
                            ),
                            3,
                        ),
                    ),
                ),
                nd.make_synset_node(
                    id=2,
                    definition=u'Duży ssak',
                    is_artificial=True,
                ),
                nd.make_synset_node(
                    id=3,
                    definition=u'Duży kundel',
                ),
                nd.make_lexical_unit_node(
                    id=11,
                    lemma=u'psisko',
                    pos=en.PoS.n,
                    variant=1,
                    synset=1,
                    unit_index=1,
                    usage_notes=(u'pot.',),
                    examples=(u'Czarne psisko na rogu.',),
                    examples_sources=(u'W',),
                    domain=en.Domain.by_db_number(1),
                    related=((u'żeńskość', 12),),
                ),
                nd.make_lexical_unit_node(
                    id=12,
                    lemma=u'suczysko',
                    pos=en.PoS.n,
                    variant=1,
                    synset=1,
                    unit_index=2,
                    definition=u'Samica dużego psa',
                    domain=en.Domain.by_db_number(1),
                ),
                nd.make_lexical_unit_node(
                    id=21,
                    lemma='ssaczysko',
                    pos=en.PoS.n,
                    variant=1,
                    synset=2,
                    unit_index=1,
                    domain=en.Domain.by_db_number(1),
                ),
                nd.make_lexical_unit_node(
                    id=31,
                    lemma=u'kundlisko',
                    pos=en.PoS.v,
                    variant=1,
                    synset=3,
                    unit_index=1,
                    external_links=(u'www.mieszance.pl', u'dogpedia.com/mutt'),
                    domain=en.Domain.by_db_number(2),
                    verb_aspect=en.VerbAspect.two,
                    is_emotional=True,
                    emotion_markedness=en.EmotionMarkedness.strong_negative,
                    emotion_names=(
                        en.EmotionName.radosc,
                        en.EmotionName.smutek,
                        en.EmotionName.wstret,
                    ),
                    emotion_valuations=(
                        en.EmotionValuation.dobro,
                        en.EmotionValuation.piekno,
                        en.EmotionValuation.prawda,
                    ),
                    emotion_example_1=u'bam',
                    emotion_example_2=u'bim',
                ),
            )),
        )


@_py2_only
class UnitWithNoSynsetTest(ut.TestCase):
    """Try to add a unit belonging to no synset.

    Should skip that synset.
    """

    def setUp(self):
        self.__dbm = wndbm.WnDbMockup()

        try:
            self.__setup_mock()
        except BaseException:
            self.__dbm.close()
            raise

    def __setup_mock(self):
        self.__dbm.add_synset(1)
        # A correct unit
        self.__dbm.add_lexical_unit(1, u'a', 2, 1, id_=1, domain=1)
        # An incorrect unit
        with self.__dbm.no_foreign_keys:
            self.__dbm.add_lexical_unit(None, u'b', 2, 1, id_=2, domain=2)

    def tearDown(self):
        self.__dbm.close()

    def runTest(self):
        with self.__dbm.sqlalchemy_url_file as db_file:
            nodes = _process_read_nodes(self, WNDBReader(db_file))

        self.assertEqual(
            nodes,
            set((
                nd.make_synset_node(id=1),
                nd.make_lexical_unit_node(
                    id=1,
                    lemma=u'a',
                    pos=en.PoS.n,
                    variant=1,
                    synset=1,
                    unit_index=0,
                    domain=en.Domain.bhp,
                ),
            )),
        )


@_py2_only
class UnitWithNonexistentSynsetTest(ut.TestCase):
    """Unit has a synset that does not exist."""

    def setUp(self):
        self.__dbm = wndbm.WnDbMockup()

        try:
            self.__setup_mock()
        except BaseException:
            self.__dbm.close()
            raise

    def __setup_mock(self):
        self.__dbm.add_synset(1)
        # A correct unit
        self.__dbm.add_lexical_unit(1, u'a', 2, 1, id_=1, domain=1)
        # An incorrect unit
        with self.__dbm.no_foreign_keys:
            self.__dbm.add_lexical_unit(2, u'b', 2, 1, id_=2, domain=1)

    def tearDown(self):
        self.__dbm.close()

    def runTest(self):
        with self.__dbm.sqlalchemy_url_file as db_file:
            nodes = _process_read_nodes(self, WNDBReader(db_file))

        self.assertEqual(
            nodes,
            set((
                nd.make_synset_node(id=1),
                nd.make_lexical_unit_node(
                    id=1,
                    lemma=u'a',
                    pos=en.PoS.n,
                    variant=1,
                    synset=1,
                    unit_index=0,
                    domain=en.Domain.by_db_number(1),
                ),
            )),
        )


@_py2_only
class SynsetWithNoUnitTest(ut.TestCase):
    """An empty synset."""

    def setUp(self):
        self.__dbm = wndbm.WnDbMockup()

        try:
            self.__setup_mock()
        except BaseException:
            self.__dbm.close()
            raise

    def __setup_mock(self):
        self.__dbm.add_synset(1)
        self.__dbm.add_synset(2)
        # One synset is going to be correct
        self.__dbm.add_lexical_unit(1, u'a', 2, id_=1, domain=1)

    def tearDown(self):
        self.__dbm.close()

    def runTest(self):
        # The empty synset should be skipped
        with self.__dbm.sqlalchemy_url_file as db_file:
            nodes = _process_read_nodes(self, WNDBReader(db_file))

        self.assertEqual(
            nodes,
            set((
                nd.make_synset_node(id=1),
                nd.make_lexical_unit_node(
                    id=1,
                    lemma=u'a',
                    pos=en.PoS.n,
                    variant=1,
                    synset=1,
                    unit_index=0,
                    domain=en.Domain.by_db_number(1),
                ),
            )),
        )


@_py2_only
class NonexistentSynsetInRelTest(ut.TestCase):
    """Try to add a relation to a bogus synset."""

    def setUp(self):
        self.__dbm = wndbm.WnDbMockup()

        try:
            self.__setup_mock()
        except BaseException:
            self.__dbm.close()
            raise

    def __setup_mock(self):
        self.__dbm.add_synset(1)
        # A synset needs a unit
        self.__dbm.add_lexical_unit(1, u'a', 2, id_=1, domain=1)
        # A relation is needed, any one
        self.__dbm.add_relation_type(u'rel', u'r', id_=1)
        # Now, relate to a synset which does not exist
        with self.__dbm.no_foreign_keys:
            self.__dbm.add_synset_relation(1, 2, 1)

    def tearDown(self):
        self.__dbm.close()

    def runTest(self):
        # Relation should be omitted, but synset and unit kept
        with self.__dbm.sqlalchemy_url_file as db_file:
            nodes = _process_read_nodes(self, WNDBReader(db_file))

        self.assertEqual(
            nodes,
            set((
                nd.make_synset_node(id=1),
                nd.make_lexical_unit_node(
                    id=1,
                    lemma=u'a',
                    pos=en.PoS.n,
                    variant=1,
                    synset=1,
                    unit_index=0,
                    domain=en.Domain.by_db_number(1),
                ),
            )),
        )


@_py2_only
class NonexistentUnitInRelTest(ut.TestCase):
    """Try to add a lexical relation to a bogus unit."""

    def setUp(self):
        self.__dbm = wndbm.WnDbMockup()

        try:
            self.__setup_mock()
        except BaseException:
            self.__dbm.close()
            raise

    def __setup_mock(self):
        self.__dbm.add_synset(1)
        self.__dbm.add_lexical_unit(1, u'a', 2, id_=1, domain=1)
        # Now, use any relation to link to nonexisting unit
        self.__dbm.add_relation_type(u'rel', u'r', is_syn=False, id_=1)
        with self.__dbm.no_foreign_keys:
            self.__dbm.add_lexical_relation(1, 2, 1)

    def tearDown(self):
        self.__dbm.close()

    def runTest(self):
        # As with synset, relation should be ignored, but unit and synset
        # should be kept
        with self.__dbm.sqlalchemy_url_file as db_file:
            nodes = _process_read_nodes(self, WNDBReader(db_file))

        self.assertEqual(
            nodes,
            set((
                nd.make_synset_node(id=1),
                nd.make_lexical_unit_node(
                    id=1,
                    lemma=u'a',
                    pos=en.PoS.n,
                    variant=1,
                    synset=1,
                    unit_index=0,
                    domain=en.Domain.by_db_number(1),
                ),
            )),
        )


@_py2_only
class NonexistentRelationTest(ut.TestCase):
    """Synsets / units exist, but the relation between them is undefined."""

    # XXX Only synset relation is checked
    def setUp(self):
        self.__dbm = wndbm.WnDbMockup()

        try:
            self.__setup_mock()
        except BaseException:
            self.__dbm.close()
            raise

    def __setup_mock(self):
        self.__dbm.add_synset(1)
        self.__dbm.add_synset(2)
        self.__dbm.add_lexical_unit(1, u'a', 2, id_=1, domain=1)
        self.__dbm.add_lexical_unit(2, u'b', 2, id_=2, domain=1)
        # Now, an undefined relation
        with self.__dbm.no_foreign_keys:
            self.__dbm.add_synset_relation(1, 2, 1)

    def tearDown(self):
        self.__dbm.close()

    def runTest(self):
        # Both synsets and units should be retained
        with self.__dbm.sqlalchemy_url_file as db_file:
            nodes = _process_read_nodes(self, WNDBReader(db_file))

        self.assertEqual(
            nodes,
            set((
                nd.make_synset_node(id=1),
                nd.make_synset_node(id=2),
                nd.make_lexical_unit_node(
                    id=1,
                    lemma=u'a',
                    pos=en.PoS.n,
                    variant=1,
                    synset=1,
                    unit_index=0,
                    domain=en.Domain.by_db_number(1),
                ),
                nd.make_lexical_unit_node(
                    id=2,
                    lemma=u'b',
                    pos=en.PoS.n,
                    variant=1,
                    synset=2,
                    unit_index=0,
                    domain=en.Domain.by_db_number(1),
                ),
            )),
        )


@_py2_only
class UnitWithNoSynsetInRelTest(ut.TestCase):
    """Unit belongs to no synset and appears in an lexical relation."""

    def setUp(self):
        self.__dbm = wndbm.WnDbMockup()

        try:
            self.__setup_mock()
        except BaseException:
            self.__dbm.close()
            raise

    def __setup_mock(self):
        self.__dbm.add_synset(1)
        self.__dbm.add_lexical_unit(1, u'a', 2, id_=1, domain=1)
        # Now, use any relation to link to a unit created with no synset
        self.__dbm.add_relation_type(u'rel', u'r', is_syn=False, id_=1)
        with self.__dbm.no_foreign_keys:
            self.__dbm.add_lexical_unit(None, u'b', 2, id_=2, domain=1)
            self.__dbm.add_lexical_relation(1, 2, 1)

    def tearDown(self):
        self.__dbm.close()

    def runTest(self):
        # Both the relation and the bogus unit should be ignored
        with self.__dbm.sqlalchemy_url_file as db_file:
            nodes = _process_read_nodes(self, WNDBReader(db_file))

        self.assertEqual(
            nodes,
            set((
                nd.make_synset_node(id=1),
                nd.make_lexical_unit_node(
                    id=1,
                    lemma=u'a',
                    pos=en.PoS.n,
                    variant=1,
                    synset=1,
                    unit_index=0,
                    domain=en.Domain.by_db_number(1),
                ),
            )),
        )


@_py2_only
class SynsetWithNoUnitInRelTest(ut.TestCase):
    """A synset in relation exists, but has no synsets."""

    def setUp(self):
        self.__dbm = wndbm.WnDbMockup()

        try:
            self.__setup_mock()
        except BaseException:
            self.__dbm.close()
            raise

    def __setup_mock(self):
        self.__dbm.add_synset(1)
        # A synset needs a unit
        self.__dbm.add_lexical_unit(1, u'a', 2, id_=1, domain=1)
        # A relation is needed, any one
        self.__dbm.add_relation_type(u'rel', u'r', id_=1)
        # Now, create an empty synset and relate to it
        self.__dbm.add_synset(2)
        with self.__dbm.no_foreign_keys:
            self.__dbm.add_synset_relation(1, 2, 1)

    def tearDown(self):
        self.__dbm.close()

    def runTest(self):
        # Relation should be omitted, as well as the empty synset, but synset
        # and unit kept
        with self.__dbm.sqlalchemy_url_file as db_file:
            nodes = _process_read_nodes(self, WNDBReader(db_file))

        self.assertEqual(
            nodes,
            set((
                nd.make_synset_node(id=1),
                nd.make_lexical_unit_node(
                    id=1,
                    lemma=u'a',
                    pos=en.PoS.n,
                    variant=1,
                    synset=1,
                    unit_index=0,
                    domain=en.Domain.by_db_number(1),
                ),
            )),
        )


def setUpModule():
    logging.getLogger('wncomments').addHandler(logging.StreamHandler())


def _process_read_nodes(utest, nodeiter):
    type_order = []
    all_nodes = set()

    for k, g in itt.groupby(nodeiter, type):
        type_order.append(k)
        all_nodes.update(g)

    # Test the order, and return the nodes for further testing.
    # The order must be: relation types, lexical units and synsets, with no
    # breaks. The first element (relation types) is optional, so don't count it
    # if the order list is shorter.
    utest.assertEqual(
        type_order,
        _NODE_ORDER[1:] if len(type_order) < 3 else _NODE_ORDER,
    )
    return all_nodes
